"""Test cases for FlotorchADKSession class."""

import time
from unittest.mock import patch

import pytest
from google.adk.events.event import Event
from google.adk.events.event_actions import EventActions
from google.adk.sessions.session import Session
from google.genai import types

from flotorch.adk.tests.test_data.session_test_data import (
    CREATE_SESSION_TEST_DATA,
    DELETE_SESSION_TEST_DATA,
    GET_SESSION_TEST_DATA,
    LIST_SESSIONS_TEST_DATA,
    APPEND_EVENT_TEST_DATA,
)


class TestFlotorchADKSessionInit:
    """Test constructor initialization."""
    
    @patch('flotorch.adk.sessions.FlotorchSession')
    def test_init_with_custom_parameters(self, mock_flotorch_session):
        """Test initialization with custom parameters."""
        from flotorch.adk.sessions import FlotorchADKSession
        
        instance = FlotorchADKSession(
            api_key="custom-key-123",
            base_url="https://custom.flotorch.com"
        )
               
        assert instance._flotorch_session is not None
        assert isinstance(instance.user_state, dict)
        assert isinstance(instance.app_state, dict)
               
        mock_flotorch_session.assert_called_once_with(
            api_key="custom-key-123",
            base_url="https://custom.flotorch.com"
        )
    
    @patch('flotorch.adk.sessions.FlotorchSession')
    def test_init_with_defaults(self, mock_flotorch_session):
        """Test initialization with required parameters."""
        from flotorch.adk.sessions import FlotorchADKSession
        
        instance = FlotorchADKSession(
            api_key="test-key",
            base_url="http://test.com"
        )
        
        assert instance._flotorch_session is not None
        assert isinstance(instance.user_state, dict)
        assert isinstance(instance.app_state, dict)
        
        mock_flotorch_session.assert_called_once_with(
            api_key="test-key",
            base_url="http://test.com"
        )


class TestFlotorchADKSessionCreate:
    """Test create_session function."""
    
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "test_name,session_id,state,should_fail,expected_exception",
        CREATE_SESSION_TEST_DATA
    )
    async def test_create_session_parametrized(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data,
        test_name,
        session_id,
        state,
        should_fail,
        expected_exception
    ):
        """Test create_session with various scenarios."""
        if should_fail:
            mock_sdk_session.create.side_effect = expected_exception("Create failed")
            with pytest.raises(expected_exception):
                await adk_session_with_mock.create_session(
                    app_name=test_data["app_name"],
                    user_id=test_data["user_id"],
                    session_id=session_id,
                    state=state
                )
        else:
            
            mock_sdk_session.create.return_value = {
                "uid": session_id or "auto-generated-id",
                "appName": test_data["app_name"],
                "userId": test_data["user_id"],
                "state": state,
                "last_update_time": 1234567890
            }
            
            session = await adk_session_with_mock.create_session(
                app_name=test_data["app_name"],
                user_id=test_data["user_id"],
                session_id=session_id,
                state=state
            )
            
            assert isinstance(session, Session)
            assert session.app_name == test_data["app_name"]
            mock_sdk_session.create.assert_called_once()
    
    def test_create_session_generates_uuid_when_none(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data
    ):
        """Test create_session generates UUID when session_id is None."""
        
        mock_sdk_session.create.return_value = {
            "uid": "auto-gen-uuid",
            "appName": test_data["app_name"],
            "userId": test_data["user_id"],
            "state": {},
            "last_update_time": 1234567890
        }
        
        session = adk_session_with_mock.create_session_sync(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"]
        )
        
        assert session.id is not None
        assert len(session.id) == 36  
    
    @pytest.mark.asyncio
    async def test_create_session_with_empty_state(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data
    ):
        """Test create_session with empty state dict."""
        mock_sdk_session.create.return_value = {
            "uid": "session-empty",
            "appName": test_data["app_name"],
            "userId": test_data["user_id"],
            "state": {},
            "last_update_time": 1234567890
        }
        
        session = await adk_session_with_mock.create_session(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"],
            state={}
        )
        
        assert session.state == {}
        mock_sdk_session.create.assert_called_once()


class TestFlotorchADKSessionGet:
    """Test get_session function."""
    
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "test_name,session_exists,events_count,has_state,should_fail,expected_exception",
        GET_SESSION_TEST_DATA
    )
    async def test_get_session_parametrized(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data,
        test_name,
        session_exists,
        events_count,
        has_state,
        should_fail,
        expected_exception
    ):
        """Test get_session with various scenarios."""
        if should_fail:
            mock_sdk_session.get.side_effect = expected_exception("Get failed")
            with pytest.raises(expected_exception):
                await adk_session_with_mock.get_session(
                    app_name=test_data["app_name"],
                    user_id=test_data["user_id"],
                    session_id=test_data["session_id"]
                )
        elif not session_exists:
            mock_sdk_session.get.return_value = None
            
            session = await adk_session_with_mock.get_session(
                app_name=test_data["app_name"],
                user_id=test_data["user_id"],
                session_id=test_data["session_id"]
            )
            
            assert session is None
        else:
            state = {"key": "value"} if has_state else {}
            
            mock_response = {
                "uid": test_data["session_id"],
                "appName": test_data["app_name"],
                "userId": test_data["user_id"],
                "state": state,
                "last_update_time": 1234567890
            }
            if events_count > 0:
                mock_response["events"] = [
                    {
                        "uid_event": f"e{i}",
                        "invocation_id": f"inv{i}",
                        "author": "user",
                        "timestamp": 1234567890.0 + i,
                        "content": {"role": "user", "parts": [{"text": f"message {i}"}]},
                        "actions": {}
                    }
                    for i in range(events_count)
                ]
            mock_sdk_session.get.return_value = mock_response
            
            session = await adk_session_with_mock.get_session(
                app_name=test_data["app_name"],
                user_id=test_data["user_id"],
                session_id=test_data["session_id"]
            )
            
            assert isinstance(session, Session)
            assert len(session.events) == events_count
    
    @pytest.mark.asyncio
    async def test_get_session_returns_events_correctly(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data
    ):
        """Test get_session correctly returns events."""
        # Create mock data directly (like Langgraph does)
        mock_sdk_session.get.return_value = {
            "uid": test_data["session_id"],
            "appName": test_data["app_name"],
            "userId": test_data["user_id"],
            "state": {},
            "last_update_time": 1234567890,
            "events": [
                {
                    "uid_event": f"e{i}",
                    "invocation_id": f"inv{i}",
                    "author": "user",
                    "timestamp": 1234567890.0 + i,
                    "content": {"role": "user", "parts": [{"text": f"message {i}"}]},
                    "actions": {}
                }
                for i in range(3)
            ]
        }
        
        session = await adk_session_with_mock.get_session(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"],
            session_id=test_data["session_id"]
        )
        
        assert len(session.events) == 3
        mock_sdk_session.get.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_get_session_with_state(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data
    ):
        """Test get_session returns session with state."""
        test_state = {"key1": "value1", "key2": "value2"}
        # Create mock data directly (like Langgraph does)
        mock_sdk_session.get.return_value = {
            "uid": test_data["session_id"],
            "appName": test_data["app_name"],
            "userId": test_data["user_id"],
            "state": test_state,
            "last_update_time": 1234567890
        }
        
        session = await adk_session_with_mock.get_session(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"],
            session_id=test_data["session_id"]
        )
        
        assert session.state == test_state


class TestFlotorchADKSessionList:
    """Test list_sessions function."""
    
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "test_name,sessions_count,should_fail,expected_exception",
        LIST_SESSIONS_TEST_DATA
    )
    async def test_list_sessions_parametrized(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data,
        test_name,
        sessions_count,
        should_fail,
        expected_exception
    ):
        """Test list_sessions with various scenarios."""
        if should_fail:
            mock_sdk_session.list.side_effect = expected_exception("List failed")
            with pytest.raises(expected_exception):
                await adk_session_with_mock.list_sessions(
                    app_name=test_data["app_name"],
                    user_id=test_data["user_id"]
                )
        else:
            
            mock_sdk_session.list.return_value = [
                {
                    "uid": f"session-{i}",
                    "appName": test_data["app_name"],
                    "userId": test_data["user_id"],
                    "state": {"index": i},
                    "last_update_time": 1234567890 + i
                }
                for i in range(sessions_count)
            ]
            
            response = await adk_session_with_mock.list_sessions(
                app_name=test_data["app_name"],
                user_id=test_data["user_id"]
            )
            
            assert response is not None
            assert len(response.sessions) == sessions_count
    
    @pytest.mark.asyncio
    async def test_list_sessions_returns_empty(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data
    ):
        """Test list_sessions returns empty list when no sessions exist."""
        mock_sdk_session.list.return_value = []
        
        response = await adk_session_with_mock.list_sessions(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"]
        )
        
        assert len(response.sessions) == 0
        mock_sdk_session.list.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_list_sessions_returns_correct_count(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data
    ):
        """Test list_sessions returns correct number of sessions."""
        
        mock_sdk_session.list.return_value = [
            {
                "uid": f"session-{i}",
                "appName": test_data["app_name"],
                "userId": test_data["user_id"],
                "state": {"index": i},
                "last_update_time": 1234567890 + i
            }
            for i in range(5)
        ]
        
        response = await adk_session_with_mock.list_sessions(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"]
        )
        
        assert len(response.sessions) == 5
        assert all(isinstance(s, Session) for s in response.sessions)


class TestFlotorchADKSessionDelete:
    """Test delete_session function."""
    
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "test_name,session_exists,should_fail,expected_exception",
        DELETE_SESSION_TEST_DATA
    )
    async def test_delete_session_parametrized(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data,
        test_name,
        session_exists,
        should_fail,
        expected_exception
    ):
        """Test delete_session with various scenarios."""
        if not session_exists:
            mock_sdk_session.get.return_value = None
            
            await adk_session_with_mock.delete_session(
                app_name=test_data["app_name"],
                user_id=test_data["user_id"],
                session_id=test_data["session_id"]
            )
            
            mock_sdk_session.get.assert_called_once()
        elif should_fail:
            mock_sdk_session.get.return_value = {
                "uid": test_data["session_id"],
                "state": {},
                "last_update_time": 1
            }
            mock_sdk_session.delete.side_effect = expected_exception("Delete failed")
            
            with pytest.raises(expected_exception):
                await adk_session_with_mock.delete_session(
                    app_name=test_data["app_name"],
                    user_id=test_data["user_id"],
                    session_id=test_data["session_id"]
                )
        else:
            mock_sdk_session.get.return_value = {
                "uid": test_data["session_id"],
                "state": {},
                "last_update_time": 1
            }
            mock_sdk_session.delete.return_value = {"status": "deleted"}
            
            await adk_session_with_mock.delete_session(
                app_name=test_data["app_name"],
                user_id=test_data["user_id"],
                session_id=test_data["session_id"]
            )
            
            mock_sdk_session.delete.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_delete_session_verifies_existence_first(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data
    ):
        """Test delete_session checks if session exists before deleting."""
        mock_sdk_session.get.return_value = {
            "uid": test_data["session_id"],
            "state": {},
            "last_update_time": 1
        }
        mock_sdk_session.delete.return_value = {"status": "deleted"}
        
        await adk_session_with_mock.delete_session(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"],
            session_id=test_data["session_id"]
        )
        
        
        assert mock_sdk_session.get.call_count == 1
        assert mock_sdk_session.delete.call_count == 1


class TestFlotorchADKSessionAppendEvent:
    """Test append_event function."""
    
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "test_name,session_exists,has_content,has_state_delta,state_delta,should_fail,expected_exception",
        APPEND_EVENT_TEST_DATA
    )
    async def test_append_event_parametrized(
        self,
        adk_session_with_mock,
        mock_sdk_session,
        test_data,
        test_name,
        session_exists,
        has_content,
        has_state_delta,
        state_delta,
        should_fail,
        expected_exception
    ):
        """Test append_event with various scenarios."""
        if not session_exists:
            mock_sdk_session.get.return_value = None
        else:
            mock_sdk_session.get.return_value = {
                "uid": test_data["session_id"],
                "state": {},
                "last_update_time": 1
            }
        
        if should_fail:
            mock_sdk_session.add_event.side_effect = expected_exception("Add event failed")
        else:
            
            mock_sdk_session.add_event.return_value = {
                "uid": "e1",
                "invocationId": "inv1",
                "author": "user"
            }
        
        
        session = Session(
            app_name=test_data["app_name"],
            user_id=test_data["user_id"],
            id=test_data["session_id"],
            state={},
            last_update_time=1
        )
        
        event_kwargs = {
            "id": "e1",
            "invocation_id": "inv",
            "author": "user",
            "timestamp": time.time()
        }
        
        if has_content:
            event_kwargs["content"] = types.Content(
                role="user",
                parts=[types.Part(text="Test message")]
            )
        
        if has_state_delta:
            event_kwargs["actions"] = EventActions(state_delta=state_delta)
        
        event = Event(**event_kwargs)
        
        if should_fail:
            with pytest.raises(expected_exception):
                await adk_session_with_mock.append_event(session, event)
        else:
            returned_event = await adk_session_with_mock.append_event(session, event)
            
            assert returned_event.id == "e1"
            
            if has_state_delta:
                for key, value in state_delta.items():
                    if key.startswith("app:"):
                        state_key = key[4:]
                        app_name = test_data["app_name"]
                        assert app_name in adk_session_with_mock.app_state
                        app_state = adk_session_with_mock.app_state
                        assert app_state[app_name][state_key] == value
                    elif key.startswith("user:"):
                        state_key = key[5:]
                        app_name = test_data["app_name"]
                        user_id = test_data["user_id"]
                        assert app_name in adk_session_with_mock.user_state
                        user_state = adk_session_with_mock.user_state
                        assert user_id in user_state[app_name]
                        assert user_state[app_name][user_id][state_key] == value



"""Test data for FlotorchAutogenMemory functions.

This module contains parametrized test data for query, add, and update_context
functions of the FlotorchAutogenMemory class.
"""

from autogen_core.memory import MemoryContent, MemoryMimeType

# ============================================================================
# QUERY TEST DATA
# ============================================================================

QUERY_TEST_DATA = [
    (
        "empty_results",
        "test query",
        {"data": []},
        0,  # No memories returned
        False,
        None
    ),
    (
        "single_memory_with_content",
        "what is the weather",
        {
            "object": "agent.memory.list",
            "data": [
                {
                    "id": "mem-123",
                    "content": "The weather is sunny",
                    "metadata": {"source": "adk_session"}
                }
            ]
        },
        1,
        False,
        None
    ),
    (
        "multiple_memories",
        "user preferences",
        {
            "object": "agent.memory.list",
            "data": [
                {"id": "mem-1", "content": "User likes Python"},
                {"id": "mem-2", "content": "User prefers dark mode"},
                {"id": "mem-3", "content": "User timezone is PST"}
            ]
        },
        3,
        False,
        None
    ),
    (
        "memory_without_content_field",
        "test",
        {
            "data": [
                {"id": "mem-1", "content": "Valid memory"},
                {"id": "mem-2", "other_field": "Invalid"},
                {"id": "mem-3", "content": ""}  # Empty content
            ]
        },
        1,  # Only one valid
        False,
        None
    ),
    (
        "api_failure",
        "test query",
        None,  # Will trigger exception
        0,
        True,
        Exception
    ),
]

# ============================================================================
# ADD TEST DATA
# ============================================================================

ADD_TEST_DATA = [
    (
        "simple_text_content",
        MemoryContent(content="User said hello", mime_type=MemoryMimeType.TEXT),
        False,
        None
    ),
    (
        "long_content",
        MemoryContent(
            content="This is a very long memory content that contains multiple sentences and a lot of information " * 10,
            mime_type=MemoryMimeType.TEXT
        ),
        False,
        None
    ),
    (
        "empty_content",
        MemoryContent(content="", mime_type=MemoryMimeType.TEXT),
        False,
        None  # Should not fail, just stores empty
    ),
    (
        "special_characters_content",
        MemoryContent(
            content="Content with special chars: @#$%^&*(){}[]<>?/\\|~`",
            mime_type=MemoryMimeType.TEXT
        ),
        False,
        None
    ),
    (
        "api_failure",
        MemoryContent(content="Test", mime_type=MemoryMimeType.TEXT),
        True,
        Exception
    ),
]

# ============================================================================
# UPDATE_CONTEXT TEST DATA
# ============================================================================

UPDATE_CONTEXT_TEST_DATA = [
    (
        "empty_messages",
        [],
        {"data": []},
        False,  # No context update
        False,
        None
    ),
    (
        "single_message_with_memories",
        ["What is Python?"],
        {
            "data": [
                {"id": "mem-1", "content": "Python is a programming language"}
            ]
        },
        True,  # Should add to context
        False,
        None
    ),
    (
        "multiple_messages_uses_last",
        ["First message", "Second message", "What is AI?"],
        {
            "data": [
                {"id": "mem-1", "content": "AI is artificial intelligence"}
            ]
        },
        True,
        False,
        None
    ),
    (
        "no_memories_found",
        ["Test query"],
        {"data": []},
        False,  # "No memories found" shouldn't be added to context
        False,
        None
    ),
    (
        "multiple_memories_formatted",
        ["Tell me about the user"],
        {
            "data": [
                {"id": "mem-1", "content": "User name is John"},
                {"id": "mem-2", "content": "User likes coding"},
                {"id": "mem-3", "content": "User is from NYC"}
            ]
        },
        True,
        False,
        None
    ),
]

# ============================================================================
# DEDUPLICATION TEST DATA
# ============================================================================

DEDUPLICATION_TEST_DATA = [
    (
        "same_content_same_hash",
        [
            MemoryContent(content="Memory A", mime_type=MemoryMimeType.TEXT),
            MemoryContent(content="Memory B", mime_type=MemoryMimeType.TEXT)
        ],
        [
            MemoryContent(content="Memory A", mime_type=MemoryMimeType.TEXT),
            MemoryContent(content="Memory B", mime_type=MemoryMimeType.TEXT)
        ],
        True  # Should be same hash
    ),
    (
        "different_content_different_hash",
        [
            MemoryContent(content="Memory A", mime_type=MemoryMimeType.TEXT)
        ],
        [
            MemoryContent(content="Memory B", mime_type=MemoryMimeType.TEXT)
        ],
        False  # Should be different hash
    ),
    (
        "different_order_same_hash",
        [
            MemoryContent(content="Memory A", mime_type=MemoryMimeType.TEXT),
            MemoryContent(content="Memory B", mime_type=MemoryMimeType.TEXT)
        ],
        [
            MemoryContent(content="Memory B", mime_type=MemoryMimeType.TEXT),
            MemoryContent(content="Memory A", mime_type=MemoryMimeType.TEXT)
        ],
        True  # Sorted before hashing, so same hash
    ),
    (
        "empty_results",
        [],
        [],
        True  # Both empty should have same hash
    ),
]


import asyncio
import os
import time
import uuid
import weakref
from collections.abc import Sequence
from typing import Any, Dict, List, Optional

# Global registry to keep track of ToolsProxy instances and their managers
_TOOLS_PROXY_REGISTRY = {}

from langchain.agents import Tool, create_openai_functions_agent
from langchain.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool
from pydantic import Field, create_model

from langchain.output_parsers import PydanticOutputParser

from langchain_mcp_adapters.client import MultiServerMCPClient

from flotorch.langchain.llm import FlotorchLangChainLLM
from flotorch.sdk.utils.http_utils import http_get
from flotorch.sdk.utils.logging_utils import (
    log_error,
    log_object_creation,
    log_warning,
    log_info
)
from flotorch.sdk.utils.validation_utils import validate_data_against_schema
from dotenv import load_dotenv
load_dotenv()

def sanitize_agent_name(name: str) -> str:
    """
    Sanitize agent name to be a valid identifier.
    
    Replaces invalid characters with underscores and ensures it starts with a 
    letter or underscore.
    """
    import re
    
    # Replace invalid characters with underscores
    sanitized = re.sub(r'[^a-zA-Z0-9_]', '_', name)
    
    # Ensure it starts with a letter or underscore
    if sanitized and not sanitized[0].isalpha() and sanitized[0] != '_':
        sanitized = f"agent_{sanitized}"
    
    # Remove consecutive underscores
    sanitized = re.sub(r'_+', '_', sanitized)
    
    # Remove leading/trailing underscores
    sanitized = sanitized.strip('_')
    
    # Ensure it's not empty
    if not sanitized:
        sanitized = "agent"
    
    return sanitized


def schema_to_pydantic_model(name: str, schema: Dict[str, Any]):
    """
    Create a minimal Pydantic model from a JSON schema dict.

    Only supports primitive field types used in simple structured outputs.
    """
    properties = schema.get("properties", {}) or {}
    fields: Dict[str, Any] = {}
    for prop, prop_schema in properties.items():
        field_type = str
        t = (prop_schema or {}).get("type")
        if t == "integer":
            field_type = int
        elif t == "number":
            field_type = float
        elif t == "boolean":
            field_type = bool
        description = (prop_schema or {}).get("description", "")
        fields[prop] = (field_type, Field(description=description))
    return create_model(name, **fields)


class FlotorchLangChainAgent:
    """
    Flotorch LangChain Agent manager that builds and returns only a LangChain agent
    using create_openai_functions_agent. No AgentExecutor or external memory/session.
    """

    def __init__(
        self,
        agent_name: str,
        enable_memory: bool = False,
        custom_tools: Optional[List[Tool]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        self.agent_name = agent_name
        self.enable_memory = enable_memory
        self.custom_tools = custom_tools or []
        self.base_url = base_url or os.environ.get("FLOTORCH_BASE_URL")
        self.api_key = api_key or os.environ.get("FLOTORCH_API_KEY")

        # Holds the currently active tool list used to build the agent
        self._tools: List[Tool] = []

        self.config = self._fetch_agent_config(agent_name)
        self._agent = asyncio.run(self._build_agent_from_config(self.config))
        self._last_reload = time.time()

        # Log object creation with key parameters (sensitive values are masked)
        log_object_creation(
            "FlotorchLangChainAgent",
            agent_name=agent_name,
            base_url=self.base_url,
            has_custom_tools=bool(custom_tools),
            sync_enabled=self.config.get("syncEnabled", False),
        )

    def _fetch_agent_config(self, agent_name: str) -> Dict[str, Any]:
        if not self.base_url:
            raise ValueError("base_url is required")
        if not self.api_key:
            raise ValueError("api_key is required")
        url = f"{self.base_url.rstrip('/')}/v1/agents/{agent_name}"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json"
        }
        return http_get(url, headers=headers)

    # ---------------------- MCP TOOL BUILDER -------------------
    def _make_tool_sync(self, tool):
        """Make async MCP tool compatible with LangGraph's sync execution."""
        from langchain_core.tools import StructuredTool
        import asyncio
        
        def sync_wrapper(**kwargs):
            try:
                result = asyncio.run(tool.ainvoke(kwargs))
                return result
            except Exception as e:
                raise
        
        return StructuredTool(
            name=tool.name,
            description=tool.description,
            args_schema=tool.args_schema,
            func=sync_wrapper
        )
        
    async def _build_mcp_tools(self, tool_cfg) -> List[BaseTool]:
        """Build MCP tools using MultiServerMCPClient."""
        try:
            cfg = tool_cfg.get("config", {})
            tool_name = tool_cfg.get("name", "mcp_tool")
            headers = dict(cfg.get("headers", {}))
            transport = cfg.get("transport", "streamable_http")
            proxy_url = f"{self.base_url}/v1/mcps/{tool_cfg['name']}/proxy"  

            if not proxy_url:
                return []

            if self.api_key:
                headers["Authorization"] = f"Bearer {self.api_key}"

            if transport == "HTTP_STREAMABLE":
                mcp_transport = "streamable_http"
            elif transport == "HTTP_SSE":
                mcp_transport = "sse"
            else:
                return []

            mcp_server_config = {
                tool_name: {
                    "url": proxy_url,
                    "transport": mcp_transport,
                }
            }

            if headers:
                mcp_server_config[tool_name]["headers"] = headers

            client = MultiServerMCPClient(mcp_server_config)
            
            if not hasattr(self, '_mcp_clients'):
                self._mcp_clients = {}
            self._mcp_clients[tool_name] = client
            
            tools = await client.get_tools(server_name=tool_name)
            
            filtered_tools = self._filter_mcp_tools(tools, tool_cfg)
            
            wrapped_tools = []
            for tool in filtered_tools:
                wrapped_tool = self._make_tool_sync(tool)
                wrapped_tools.append(wrapped_tool)
            
            return wrapped_tools

        except Exception as e:
            log_error("FlotorchLangChainAgent._build_mcp_tools", e)
            return []


    def _filter_mcp_tools(self, all_tools: List[BaseTool], tool_cfg) -> List[BaseTool]:
        """Filter MCP tools based on configuration."""
        filtered_tools = []
        tool_name = sanitize_agent_name(tool_cfg.get("name"))

        if tool_name:
            # Match tool name to config (avoid collisions)
            for tool in all_tools:
                if getattr(tool, 'name', None) == tool_name:
                    filtered_tools.append(tool)
        else:
            # If no specific name, use all tools
            filtered_tools = all_tools

        return filtered_tools


    async def _build_tools(self, config) -> List[BaseTool]:
        tools = []
        # Custom tools always included
        if self.custom_tools:
            tools.extend(self.custom_tools)

        # Build MCP tools using langchain-mcp-adapters
        # Each tool manages its own session lifecycle
        for tool_cfg in config.get("tools", []):
            if tool_cfg.get("type") == "MCP":
                
                try:
                    mcp_tools = await self._build_mcp_tools(tool_cfg)
                    tools.extend(mcp_tools)
                except Exception as e:
                    log_error("FlotorchLangChainAgent._build_tools", e)
                    log_warning(
                        f"[MCP] Skipping tool '{tool_cfg.get('name')}' due to error."
                    )

        return tools



    async def _build_agent_from_config(self, config):
        # Build prompt from config goal and systemPrompt, include input and agent_scratchpad
        messages: List[Any] = []
        goal_text = config.get("goal")
        if goal_text:
            messages.append(("system", goal_text))
        system_prompt = config.get("systemPrompt")
        if system_prompt:
            messages.append(("system", system_prompt))

        # Optional structured output via config["outputSchema"]
        output_schema = config.get("outputSchema")
        parser: Optional[PydanticOutputParser] = None
        if isinstance(output_schema, dict) and output_schema.get("properties"):
            try:
                model = schema_to_pydantic_model("OutputSchema", output_schema)
                parser = PydanticOutputParser(pydantic_object=model)
                messages.append(("system", "{format_instructions}"))
            except Exception:
                parser = None

        # Add chat history placeholder for proper message flow
        messages.append(("placeholder", "{chat_history}"))
        
        # Human input and agent scratchpad
        messages.append(("human", "{input}"))
        messages.append(("assistant","Sessions-Memory Data: {history}"))
        if self.enable_memory:
            messages.append(("assistant","Longterm-Memory Data: {longterm_history}"))
        messages.append(("placeholder", "{agent_scratchpad}"))

        prompt = ChatPromptTemplate.from_messages(messages)
        if parser is not None:
            prompt = prompt.partial(format_instructions=parser.get_format_instructions())

        llm_params = config.get("llm", {})
        model_name = llm_params.get("callableName")
        llm = FlotorchLangChainLLM(
            model_id=model_name,
            api_key=self.api_key,
            base_url=self.base_url,
        )

        tools = await self._build_tools(config)
        if not tools:
            # Ensure at least one function for create_openai_functions_agent
            try:
                def _noop_tool(input: str = "") -> str:
                    return ""

                tools = tools + [
                    Tool(
                        name="noop",
                        func=_noop_tool,
                        description="No-op tool when no tools are configured.",
                    )
                ]
            except Exception:
                pass

        # Store the tools used for this agent so callers can retrieve them
        self._tools = tools

        agent = create_openai_functions_agent(llm, tools, prompt)
        
        return agent

    def get_agent(self):
        return AgentProxy(self)

    def get_tools(self) -> List[Tool]:
        """Return the currently active tool list used to build the agent."""
        return ToolsProxy(self)

    def _get_synced_agent(self):
        sync_enabled = self.config.get('syncEnabled', False)
        if not sync_enabled:
            return self._agent

        sync_interval = self.config.get('syncInterval', 60)
        now = time.time()
        if now - self._last_reload > sync_interval:
            log_info("[Sync] Reload interval passed. Attempting to reload agent...")
            try:
                new_config = self._fetch_agent_config(self.agent_name)
                if new_config and new_config != self.config:
                    self.config = new_config
                    tools = asyncio.run(self._build_tools(self.config))
                    self._tools = tools
                    self._agent = asyncio.run(self._build_agent_from_config(self.config))
                    log_info("[Sync] Agent successfully reloaded.")
            except Exception as e:
                log_error(
                    "[Sync] Failed to reload agent config. Using previous agent.",
                    e,
                )
            finally:
                self._last_reload = now
        return self._agent


    def _get_synced_tools(self):
        sync_enabled = self.config.get('syncEnabled', False)
        if not sync_enabled:
            return self._agent

        sync_interval = self.config.get('syncInterval', 60)
        now = time.time()
        if now - self._last_reload > sync_interval:
            log_info("[Sync] Reload interval passed. Attempting to reload agent...")
            try:
                new_config = self._fetch_agent_config(self.agent_name)
                if new_config and new_config != self.config:
                    self.config = new_config
                    tools = asyncio.run(self._build_tools(self.config))
                    self._tools = tools
                    self._agent = asyncio.run(self._build_agent_from_config(self.config))
                    log_info("[Sync] Agent successfully reloaded.")
            except Exception as e:
                log_error(
                    "[Sync] Failed to reload agent config. Using previous agent.",
                    e,
                )
            finally:
                self._last_reload = now
        return self._tools

class AgentProxy(Runnable):
    def __init__(self, manager: 'FlotorchLangChainAgent'):
        self._manager = manager
    
    def __getattr__(self, item):
        return getattr(self._manager._get_synced_agent(), item)

    def __setattr__(self, key, value):
        if key == "_manager":
            return object.__setattr__(self, key, value)
        return setattr(self._manager._get_synced_agent(), key, value)

    def invoke(self, input: Any, config: Any | None = None, **kwargs: Any):
        if self._manager.config.get("inputSchema"):
            self._validate_input(input)
        return self._manager._get_synced_agent().invoke(input, config=config, **kwargs)

    async def ainvoke(self, input: Any, config: Any | None = None, **kwargs: Any):
        if self._manager.config.get("inputSchema"):
            self._validate_input(input)
        return await self._manager._get_synced_agent().ainvoke(input, config=config, **kwargs)

    def _validate_input(self, input_data: Any) -> None:
        """Validate input format and schema."""
        input_schema = self._manager.config.get("inputSchema")
        # Check if input is dict with "input" key (LangChain standard format)
        if not isinstance(input_data, dict) or "input" not in input_data:
            raise ValueError(
                f"Input validation failed: LangChain expects dict with 'input' key. "
                f"Received: {type(input_data)}"
            )

        user_data = input_data.get("input")
        is_valid = validate_data_against_schema(user_data, input_schema)

        if not is_valid:
            raise ValueError(
                f"Input validation failed: data does not match required schema. "
                f"Expected schema: {input_schema}"
            )

    def get_tools(self) -> List[Tool]:
        return ToolsProxy(self._manager)
        

class ToolsProxy(list):
    """
    A dynamic list that automatically syncs with the latest tools from a FlotorchLangChainAgent.
    
    This proxy ensures that AgentExecutor always has access to the most up-to-date tools,
    even when the agent configuration is dynamically updated.
    """

    def __init__(self, manager):
        """Initialize the ToolsProxy with a manager or list of tools."""
        super().__init__()
        
        if isinstance(manager, list):
            # Manager is already a list (corrupted case from LangChain)
            self.extend(manager)
            self._is_corrupted = True
            self._manager_ref = None
            self._try_recover_manager()
        elif hasattr(manager, '_get_synced_tools'):
            # Valid FlotorchLangChainAgent manager
            self._is_corrupted = False
            self._manager_ref = weakref.ref(manager) if manager else None
            # Store in global registry for recovery
            self._proxy_id = str(uuid.uuid4())
            _TOOLS_PROXY_REGISTRY[self._proxy_id] = manager
            self._refresh_tools()
        else:
            # Fallback for unknown manager types
            self._is_corrupted = True
            self._manager_ref = None

    def _try_recover_manager(self):
        """Try to recover manager from global registry."""
        if '_TOOLS_PROXY_REGISTRY' in globals() and _TOOLS_PROXY_REGISTRY:
            for proxy_id, manager in _TOOLS_PROXY_REGISTRY.items():
                if manager and hasattr(manager, '_get_synced_tools'):
                    self._manager_ref = weakref.ref(manager)
                    self._is_corrupted = False
                    self._proxy_id = proxy_id
                    return True
        return False

    def _refresh_tools(self):
        """Get fresh tools from the manager."""
        if self._is_corrupted or not self._manager_ref:
            # Try to recover manager from global registry if available
            try:
                proxy_id = object.__getattribute__(self, '_proxy_id')
                if '_TOOLS_PROXY_REGISTRY' in globals():
                    manager = _TOOLS_PROXY_REGISTRY.get(proxy_id)
                    if manager and hasattr(manager, '_get_synced_tools'):
                        self._manager_ref = weakref.ref(manager)
                        self._is_corrupted = False
                    else:
                        return
                else:
                    return
            except AttributeError:
                return
            
        manager = self._manager_ref()
        if manager and hasattr(manager, '_get_synced_tools'):
            synced_tools = manager._get_synced_tools()
            self.clear()
            self.extend(synced_tools)

    def __getitem__(self, index):
        """Get a tool by index, ensuring tools are up-to-date."""
        self._refresh_tools()
        return super().__getitem__(index)

    def __len__(self):
        """Get the number of tools, ensuring tools are up-to-date."""
        self._refresh_tools()
        return super().__len__()

    def __iter__(self):
        """Iterate over tools, ensuring tools are up-to-date."""
        self._refresh_tools()
        return super().__iter__()

    def __getattr__(self, name):
        """Refresh tools before any attribute access."""
        if name.startswith('_'):
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")
        self._refresh_tools()
        return super().__getattribute__(name)

    def __contains__(self, item):
        """Check if tool exists, ensuring tools are up-to-date."""
        self._refresh_tools()
        return super().__contains__(item)

    def count(self, value):
        """Count occurrences of a tool, ensuring tools are up-to-date."""
        self._refresh_tools()
        return super().count(value)

    def index(self, value, start=0, stop=None):
        """Find index of a tool, ensuring tools are up-to-date."""
        self._refresh_tools()
        return super().index(value, start, stop)


from __future__ import annotations
from typing import Any, Sequence, Iterator
import uuid

from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.base import (
    BaseCheckpointSaver,
    Checkpoint,
    CheckpointMetadata,
    CheckpointTuple,
    ChannelVersions,
)
from flotorch.sdk.utils.http_utils import APIError   
from flotorch.sdk.session import FlotorchSession

class FlotorchLanggraphSession(BaseCheckpointSaver[str]):
    
    def __init__(self, api_key: str, base_url: str, app_name: str, user_id: str, *args, **kwargs):
        """Initialize the FlotorchLanggraphSession with API credentials and configuration."""
        super().__init__(*args, **kwargs)
        self.session = FlotorchSession(api_key, base_url)
        self.app_name = app_name
        self.user_id = user_id
  
    def get_tuple(self, config: RunnableConfig) -> CheckpointTuple | None:
        uid = config["configurable"]["thread_id"]

        try:
            data = self.session.get(uid)
        except APIError as e:
            if e.status_code == 404:
                self.session.create(app_name=self.app_name, user_id=self.user_id, uid=uid)
                return None
            else:
                raise

        if not data:
            return None
            
        events = data.get("events", [])
        if not events:
            return None

        all_messages = self._extract_messages_from_events(events)
        if not all_messages:
            return None

        unique_messages = self._deduplicate_messages(all_messages, uid)
        
        checkpoint = self._create_checkpoint_structure(unique_messages)
        
        metadata = {
            "step": 0,
            "source": "input",
            "parents": {}
        }
        
        return CheckpointTuple(
            config=config,
            checkpoint=checkpoint,
            metadata=metadata,
        )

    async def aget_tuple(self, config: RunnableConfig) -> CheckpointTuple | None:
        """Async version of get_tuple."""
        return self.get_tuple(config)
  
    def put(
        self,
        config: RunnableConfig,
        checkpoint: Checkpoint,
        metadata: CheckpointMetadata,
        new_versions: ChannelVersions,
    ) -> None:

        uid = config["configurable"]["thread_id"]

        messages = checkpoint.get("messages", [])
        if not messages:
            return

        simple_messages = self._convert_messages_to_simple_format(messages)

        safe_checkpoint = {
            "messages": simple_messages,
            "v": checkpoint.get("v", 1),
            "id": checkpoint.get("id", str(uuid.uuid4())),
            "ts": checkpoint.get("ts", ""),
        }

        safe_metadata = self._sanitize_for_json(metadata)
        self.session.add_event(
            uid=uid,
            invocation_id=str(uuid.uuid4()),
            author="system",
            content={
                "metadata": safe_metadata,
                "checkpoint": safe_checkpoint,
                "new_versions": new_versions,
            }
        )

    async def aput(
        self,
        config: RunnableConfig,
        checkpoint: Checkpoint,
        metadata: CheckpointMetadata,
        new_versions: ChannelVersions,
    ) -> None:
        """Async version of put."""
        self.put(config, checkpoint, metadata, new_versions)

    def put_writes(
        self,
        config: RunnableConfig,
        writes: Sequence[tuple[str, Any]],
        task_id: str,
        task_path: str = "",
    ) -> None:
        uid = config["configurable"]["thread_id"]
        
        safe_writes = {}
        for channel, value in writes:
            safe_writes[channel] = self._sanitize_for_json(value)
        
        self.session.add_event(
            uid=uid,
            invocation_id=str(uuid.uuid4()),
            author="system",
            content={
                "writes": safe_writes,
                "task_id": task_id,
                "task_path": task_path,
            }
        )

    async def aput_writes(
        self,
        config: RunnableConfig,
        writes: Sequence[tuple[str, Any]],
        task_id: str,
        task_path: str = "",
    ) -> None:
        """Async version of put_writes."""
        self.put_writes(config, writes, task_id, task_path)

    def list(
        self,
        config: RunnableConfig | None,
        *,
        filter: dict[str, Any] | None = None,
        before: RunnableConfig | None = None,
        limit: int | None = None,
    ) -> Iterator[CheckpointTuple]:
        """List all sessions for this app/user."""
        try:
            sessions = self.session.list(app_name=self.app_name, user_id=self.user_id)
            
            count = 0
            for session_data in sessions:
                if limit and count >= limit:
                    break
                    
                session_uid = session_data.get("uid")
                if not session_uid:
                    continue

                session_config = {"configurable": {"thread_id": session_uid}}
                checkpoint = {"v": 1, "id": session_uid, "ts": "", "channel_values": {"messages": []}, "channel_versions": {}}
                metadata = {"step": 0, "source": "list", "parents": {}}
                
                yield CheckpointTuple(
                    config=session_config,
                    checkpoint=checkpoint,
                    metadata=metadata,
                )
                count += 1
                
        except APIError:
            return

    async def alist(
        self,
        config: RunnableConfig | None,
        *,
        filter: dict[str, Any] | None = None,
        before: RunnableConfig | None = None,
        limit: int | None = None,
    ) -> Iterator[CheckpointTuple]:
        """Async version of list - List all sessions for this app/user."""
        return self.list(config, filter=filter, before=before, limit=limit)

    def delete_thread(self, thread_id: str) -> None:
        try:
            self.session.delete(thread_id)
        except APIError as e:
            if e.status_code == 404:
                pass
            else:
                raise
   
    def _extract_messages_from_events(self, events: list) -> list:
        """Extract messages from event data structure."""
        all_messages = []
        
        for event in events:
            content = event.get("content", {})
            
            if "checkpoint" in content:
                messages = content["checkpoint"].get("channel_values", {}).get("messages", [])
                if messages and isinstance(messages, list):
                    all_messages.extend(messages)
                    continue
            
            if "writes" in content:
                messages = content["writes"].get("messages", [])
                if isinstance(messages, list):
                    all_messages.extend(messages)
                elif isinstance(messages, str) and messages.strip():
                    # Handle case where messages is a string (user input)
                    all_messages.append({
                        "type": "human",
                        "content": messages.strip()
                    })
        
        return all_messages

    def _deduplicate_messages(self, messages: list, uid: str = None) -> list:
        """Simple deduplication that handles tool response pairing."""
        if not messages:
            return []
        
        try:
            all_tool_calls = {}
            all_tool_responses = {}
            ai_messages = []
            other_messages = []
            
            for msg in messages:
                if not isinstance(msg, dict):
                    continue
                    
                msg_type = msg.get("type")
                
                if msg_type == "ai" and msg.get("tool_calls"):
                    ai_messages.append(msg) 
                    tool_calls = msg.get("tool_calls", [])
                    if isinstance(tool_calls, list):
                        for tool_call in tool_calls:
                            if isinstance(tool_call, dict):
                                tool_call_id = tool_call.get("id")
                                if tool_call_id and isinstance(tool_call_id, str):
                                    all_tool_calls[tool_call_id] = tool_call
                                    
                elif msg_type == "tool":
                    tool_call_id = msg.get("tool_call_id")
                    if tool_call_id:
                        all_tool_responses[tool_call_id] = msg
                else:
                    other_messages.append(msg)
    
            processed_messages = []
            
            for ai_msg in ai_messages:
                tool_calls = ai_msg.get("tool_calls", [])
                if isinstance(tool_calls, list):
                    all_responses_exist = all(
                        isinstance(tool_call, dict) and 
                        tool_call.get("id") in all_tool_responses
                        for tool_call in tool_calls
                    )
                    
                    if all_responses_exist:
                        processed_messages.append(ai_msg)
                        for tool_call in tool_calls:
                            if isinstance(tool_call, dict):
                                tool_call_id = tool_call.get("id")
                                if tool_call_id in all_tool_responses:
                                    processed_messages.append(all_tool_responses[tool_call_id])
                else:
                    processed_messages.append(ai_msg)
            
            processed_messages.extend(other_messages)
            
            return processed_messages
            
        except Exception as e:
            return messages

    def _create_checkpoint_structure(self, messages: list) -> dict:
        """Create the checkpoint structure for the session."""
        return {
            "v": 4,  
            "id": str(uuid.uuid4()),
            "ts": "",
            "channel_values": {"messages": messages},
            "channel_versions": {"messages": len(messages)},
            "updated_channels": ["messages"],
            "versions_seen": {
                "__input__": {},
                "__start__": {"__start__": 1},
                "call_model": {"branch:to:call_model": 1}
            }
        }

    def _convert_messages_to_simple_format(self, messages: list) -> list:
        """Convert messages to a simple, serializable format."""
        simple_messages = []
        
        for msg in messages:
            if isinstance(msg, dict):
                simple_messages.append(self._sanitize_for_json(msg))
            else:
                simple_messages.append(str(msg))
                
        return simple_messages

    def _sanitize_for_json(self, obj: Any) -> Any:
        """Recursively sanitize objects for JSON serialization."""
        if isinstance(obj, (list, tuple)):
            return [self._sanitize_for_json(item) for item in obj]
        elif isinstance(obj, dict):
            return {key: self._sanitize_for_json(value) for key, value in obj.items()}
        elif hasattr(obj, 'dict'):
            return obj.dict()
        else:
            return str(obj)


"""Pytest fixtures for CrewAI tests.

This module provides reusable fixtures for testing CrewAI components
including sessions, memory storage, and LLM integration.
"""

import pytest
from unittest.mock import Mock, patch
from pydantic import BaseModel, Field

from flotorch.crewai.sessions import FlotorchCrewAISession
from flotorch.crewai.memory import FlotorchMemoryStorage
from flotorch.crewai.llm import FlotorchCrewAILLM
from flotorch.crewai.tests.test_data.agent_test_data import MINIMAL_CONFIG

class MockLLMResponse:
    """Mock LLM response object for testing."""

    def __init__(self, content: str):
        """Initialize mock response.

        Args:
            content: Response content string
        """
        self.content = content


class SampleTaskSchema(BaseModel):
    """Sample Pydantic schema for task output testing."""

    result: str = Field(description="Result")
    score: float = Field(description="Score")


@pytest.fixture
def mock_session_client():
    """Create mock FlotorchSession client with default responses.

    Returns:
        Mock: Configured mock session client
    """
    mock_client = Mock()
    mock_client.create.return_value = {"uid": "test-session-123"}
    mock_client.get.return_value = {"uid": "test-session-123"}
    mock_client.add_event.return_value = {"success": True}
    mock_client.get_events.return_value = []
    mock_client.delete.return_value = None
    return mock_client


@pytest.fixture
def session_instance(mock_session_client, monkeypatch):
    """Create FlotorchCrewAISession instance with mocked client.

    Args:
        mock_session_client: Mocked session client fixture
        monkeypatch: Pytest monkeypatch fixture

    Returns:
        FlotorchCrewAISession: Instance with mocked dependencies
    """
    instance = FlotorchCrewAISession(
        base_url="https://test.flotorch.com",
        api_key="test-api-key",
        app_name="test_app",
        user_id="test_user"
    )
    monkeypatch.setattr(instance, "_session_client", mock_session_client)
    return instance


@pytest.fixture
def session_instance_with_id(session_instance):
    """Create session instance with pre-existing session_id.

    Args:
        session_instance: Base session instance fixture

    Returns:
        FlotorchCrewAISession: Instance with session_id set
    """
    session_instance._session_id = "existing-session-123"
    return session_instance


@pytest.fixture
def mock_memory_client():
    """Create mock FlotorchMemory client with default responses.

    Returns:
        Mock: Configured mock memory client
    """
    mock_client = Mock()
    mock_client.add.return_value = {"success": True}
    mock_client.search.return_value = {"data": []}
    mock_client.delete.return_value = None
    return mock_client


@pytest.fixture
def memory_instance(mock_memory_client, monkeypatch):
    """Create FlotorchMemoryStorage instance with mocked client.

    Args:
        mock_memory_client: Mocked memory client fixture
        monkeypatch: Pytest monkeypatch fixture

    Returns:
        FlotorchMemoryStorage: Instance with mocked dependencies
    """
    instance = FlotorchMemoryStorage(
        name="test_memory",
        api_key="test-api-key",
        base_url="https://test.flotorch.com",
        user_id="test_user",
        app_id="test_app"
    )
    monkeypatch.setattr(instance, "_memory", mock_memory_client)
    return instance


@pytest.fixture
def mock_task_with_output_pydantic():
    """Create mock task with output_pydantic schema.

    Returns:
        Mock: Task with output_pydantic set to SampleTaskSchema
    """
    task = Mock()
    task.output_json = None
    task.output_pydantic = SampleTaskSchema
    return task


@pytest.fixture
def mock_llm():
    """Create mock FlotorchLLM instance.

    Returns:
        Mock: FlotorchLLM mock with default invoke response
    """
    mock = Mock()
    mock.invoke.return_value = MockLLMResponse("test response")
    return mock


@pytest.fixture
def llm_instance(mock_llm):
    """Create FlotorchCrewAILLM instance with mocked FlotorchLLM.

    Args:
        mock_llm: Mocked FlotorchLLM fixture

    Returns:
        FlotorchCrewAILLM: Instance with mocked LLM dependency
    """
    with patch('flotorch.crewai.llm.FlotorchLLM', return_value=mock_llm):
        llm = FlotorchCrewAILLM(
            model_id="test-model",
            api_key="test-key",
            base_url="https://test.com"
        )
        llm.llm = mock_llm
    return llm


# Agent test fixtures

@pytest.fixture
def mock_agent_deps():
    """Mock all external dependencies for agent testing.
    
    Provides comprehensive mocking environment for FlotorchCrewAIAgent:
    - HTTP requests (config fetching)
    - CrewAI Agent and Task classes
    - FlotorchCrewAILLM
    - MCPServerAdapter (prevents connection attempts)
    - Time module (for sync testing)
    
    Returns:
        dict: Mocked objects with keys:
            - http_get: HTTP request mock
            - agent: CrewAI Agent class mock
            - task: CrewAI Task class mock
            - llm: FlotorchCrewAILLM mock
            - mcp: MCPServerAdapter mock
            - time: Time module mock
    """
    with patch('flotorch.crewai.agent.http_get') as http_mock, \
         patch('flotorch.crewai.agent.Agent') as agent_mock, \
         patch('flotorch.crewai.agent.Task') as task_mock, \
         patch('flotorch.crewai.agent.FlotorchCrewAILLM') as llm_mock, \
         patch('flotorch.crewai.agent.MCPServerAdapter') as mcp_mock, \
         patch('flotorch.crewai.agent.time') as time_mock:
        
        # Import test data here to avoid circular imports
        
        # Configure default return values
        http_mock.return_value = MINIMAL_CONFIG
        agent_mock.return_value = Mock(role="test-role", tools=[])
        task_mock.return_value = Mock(description="test-desc")
        llm_mock.return_value = Mock()
        time_mock.time.return_value = 1000.0
        
        # Prevent MCP connection attempts
        mcp_adapter = Mock(tools=[])
        mcp_mock.return_value = mcp_adapter
        
        yield {
            'http_get': http_mock,
            'agent': agent_mock,
            'task': task_mock,
            'llm': llm_mock,
            'mcp': mcp_mock,
            'time': time_mock
        }


"""LangChain-compatible session memory backed by Flotorch Sessions.

This module implements a minimal `BaseMemory` that persists user/assistant
turns as Flotorch session events and exposes a single history string under
`memory_key` for prompt formatting.
"""

from typing import Any, Dict, List, Optional
import uuid

from langchain.schema import BaseMemory
from flotorch.sdk.session import FlotorchSession
from flotorch.sdk.utils.logging_utils import (
    log_error,
    log_object_creation,
    log_warning,
)


class FlotorchLangChainSession(BaseMemory):
    """A LangChain `BaseMemory` that stores history in Flotorch Sessions.

    - Persists user/assistant turns as Flotorch session events.
    - Loads previous turns and exposes them as a single history string
      under `memory_key` for prompt formatting.

    Only `api_key` and `base_url` are required at construction time.
    `app_name` and `user_id` provide sensible defaults for standalone usage.
    """

    memory_key: str = "history"

    # Required
    api_key: Optional[str] = None
    base_url: Optional[str] = None

    # Defaults for session grouping
    app_name: str = "langchain_session_app"
    user_id: str = "langchain_user"

    # Internal client/state
    session_client: Optional[FlotorchSession] = None
    _session_id: Optional[str] = None

    def __init__(
        self,
        api_key: str,
        base_url: str,
        app_name :Optional[str] = "default_app",
        user_id: Optional[str] = "default_user"
        
    ) -> None:
        """Initialize the session memory and underlying Flotorch client.

        Parameters
        - api_key: Flotorch API key for authentication.
        - base_url: Flotorch Gateway base URL.
        - memory_key: Prompt variable name used to inject history.
        """
        memory_key  = "history"
        super().__init__(
            api_key=api_key,
            base_url=base_url,
            user_id=user_id,
            app_id=app_name,
            memory_key= memory_key
        )

        if not api_key:
            log_warning("FlotorchLangChainSession: api_key is not set.")
        if not base_url:
            log_warning("FlotorchLangChainSession: base_url is not set.")

        
        self.session_client = FlotorchSession(api_key=api_key, base_url=base_url)

        # Object creation log with relevant context.
        log_object_creation(
            "FlotorchLangChainSession",
            base_url=base_url,
            app_name=self.app_name,
            user_id=self.user_id,
            memory_key=self.memory_key,
        )

    @property
    def memory_variables(self) -> List[str]:
        """Names of variables this memory adds to the prompt.

        Returns a single-element list containing `self.memory_key`.
        """
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Load conversation history from Flotorch session events as a string."""
        try:
            session_id = self._ensure_session()
            events = self.session_client.get_events(session_id) or []
            if not events:
                log_warning(
                    "FlotorchLangChainSession.load_memory_variables: no events found."
                )
            # Ensure chronological order (oldest -> newest) for prompt readability
            events = list(events)[::-1]
            lines = self._extract_text_lines_from_events(events)
            return {self.memory_key: "\n".join(lines)}
        except Exception as e:
            log_error("FlotorchLangChainSession.load_memory_variables", e)
            return {self.memory_key: ""}

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Persist the latest user input and assistant response to the session."""
        try:
            session_id = self._ensure_session()
            user_text = inputs.get("input") or next(iter(inputs.values()))
            assistant_text = (
                outputs.get("response")
                or outputs.get("output")
                or next(iter(outputs.values()))
            )
            if not user_text:
                log_warning(
                    "FlotorchLangChainSession.save_context: empty user text."
                )
            if not assistant_text:
                log_warning(
                    "FlotorchLangChainSession.save_context: empty assistant text."
                )

            # Save user event
            self.session_client.add_event(
                uid=session_id,
                invocation_id=str(uuid.uuid4()),
                author="user",
                content={"parts": [{"text": str(user_text)}]},
            )

            # Save assistant event (mark turn complete)
            self.session_client.add_event(
                uid=session_id,
                invocation_id=str(uuid.uuid4()),
                author="assistant",
                content={"parts": [{"text": str(assistant_text)}]},
                turn_complete=True,
            )
        except Exception as e:
            # Best-effort; do not break the chain
            log_error("FlotorchLangChainSession.save_context", e)

    def clear(self) -> None:
        """Delete the active session remotely and reset local state."""
        try:
            if self._session_id:
                try:
                    self.session_client.delete(self._session_id)
                finally:
                    self._session_id = None
                log_warning(
                    "FlotorchLangChainSession.clear: session cleared and local state reset."
                )
        except Exception as e:
            # Ignore cleanup errors but log for observability
            log_error("FlotorchLangChainSession.clear", e)
            self._session_id = None

    # --- Internal helpers ---

    def _ensure_session(self) -> str:
        if self._session_id:
            try:
                data = self.session_client.get(uid=self._session_id)
                if data:
                    return self._session_id
            except Exception:
                log_warning(
                    "FlotorchLangChainSession._ensure_session: previous session invalid."
                )
                self._session_id = None

        created = self.session_client.create(
            app_name=self.app_name,
            user_id=self.user_id,
        )
        self._session_id = (
            created.get("uid")
            or created.get("id")
            or str(uuid.uuid4())
        )
        # Note: session creation is transient and frequent; avoid verbose object
        # creation logs here to prevent noise.
        return self._session_id

    @staticmethod
    def _extract_text_lines_from_events(events: List[Dict[str, Any]]) -> List[str]:
        """Extract text lines from events, prefixed with the event author if present.

        Produces lines like "user: hello" and "assistant: hi" when the
        `author` field is available; otherwise, falls back to plain text.
        """
        lines: List[str] = []
        for event in events:
            author = event.get("author")
            prefix = f"{author}: " if isinstance(author, str) and author else ""

            content = event.get("content")
            if isinstance(content, dict):
                parts = content.get("parts")
                if isinstance(parts, list):
                    for part in parts:
                        text = part.get("text") if isinstance(part, dict) else None
                        if isinstance(text, str) and text.strip():
                            lines.append(f"{prefix}{text.strip()}")
                            continue

            text_fallback = event.get("text")
            if isinstance(text_fallback, str) and text_fallback.strip():
                lines.append(f"{prefix}{text_fallback.strip()}")
        return lines





"""Test data for AutoGen agent tests.

This module provides sample configurations and mock objects for testing
the FlotorchAutogenAgent class.
"""


# Base minimal configuration for testing
MINIMAL_CONFIG = {
    "uid": "test-uid",
    "name": "test-agent",
    "version": "1",
    "systemPrompt": "Test system prompt",
    "syncEnabled": False,
    "syncInterval": 1000,
    "inputSchema": None,
    "outputSchema": None,
    "tools": [],
    "llm": {"callableName": "test-model"}
}

# Configuration with output schema for structured output testing
CONFIG_WITH_OUTPUT_SCHEMA = {
    **MINIMAL_CONFIG,
    "outputSchema": {
        "type": "object",
        "properties": {
            "result": {"type": "string"},
            "score": {"type": "number"}
        }
    }
}

# Configuration with input schema for structured input testing
CONFIG_WITH_INPUT_SCHEMA = {
    **MINIMAL_CONFIG,
    "inputSchema": {
        "type": "object",
        "properties": {
            "query": {"type": "string"},
            "context": {"type": "string"}
        }
    }
}

# Configuration with both input and output schemas
CONFIG_WITH_BOTH_SCHEMAS = {
    **CONFIG_WITH_INPUT_SCHEMA,
    **CONFIG_WITH_OUTPUT_SCHEMA
}

# Configuration with sync enabled for testing dynamic updates
CONFIG_WITH_SYNC = {
    **MINIMAL_CONFIG,
    "syncEnabled": True,
    "syncInterval": 10  # 10 seconds for testing
}

# Modified configuration for testing config change detection
CONFIG_MODIFIED = {
    **CONFIG_WITH_SYNC,
    "systemPrompt": "Modified system prompt",
    "version": "2"
}

# Configuration with MCP tools
CONFIG_WITH_MCP_TOOLS = {
    **MINIMAL_CONFIG,
    "tools": [
        {
            "type": "MCP",
            "name": "test-mcp-sse",
            "config": {
                "transport": "HTTP_SSE",
                "headers": {}
            }
        },
        {
            "type": "MCP",
            "name": "test-mcp-stream",
            "config": {
                "transport": "HTTP_STREAMABLE",
                "headers": {}
            }
        }
    ]
}

# Test data for schema conversion
SCHEMA_TO_PYDANTIC_DATA = [
    {
        "id": "single_property_input",
        "name": "InputSchema",
        "schema": {
            "type": "object",
            "properties": {
                "query": {"type": "string", "description": "User query"}
            }
        },
        "expected_model_name": "QueryInput"
    },
    {
        "id": "single_property_output",
        "name": "OutputSchema",
        "schema": {
            "type": "object",
            "properties": {
                "result": {"type": "string", "description": "Result"}
            }
        },
        "expected_model_name": "ResultOutput"
    },
    {
        "id": "multiple_properties",
        "name": "TestSchema",
        "schema": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name"},
                "age": {"type": "integer", "description": "Age"},
                "active": {"type": "boolean", "description": "Active status"}
            }
        },
        "expected_model_name": "TestSchema"
    },
    {
        "id": "number_types",
        "name": "NumberSchema",
        "schema": {
            "type": "object",
            "properties": {
                "score": {"type": "number", "description": "Score"},
                "count": {"type": "integer", "description": "Count"}
            }
        },
        "expected_model_name": "NumberSchema"
    }
]

# Test data for agent initialization
AGENT_INIT_DATA = [
    {
        "id": "basic_init",
        "params": {
            "agent_name": "test-agent",
            "base_url": "https://test.com",
            "api_key": "test-key"
        },
        "expected": {
            "agent_name": "test-agent",
            "base_url": "https://test.com",
            "api_key": "test-key"
        }
    },
    {
        "id": "with_memory",
        "params": {
            "agent_name": "test-agent",
            "memory": ["mock_memory"],
            "base_url": "https://test.com",
            "api_key": "test-key"
        },
        "expected": {
            "agent_name": "test-agent",
            "memory": ["mock_memory"]
        }
    },
    {
        "id": "with_custom_tools",
        "params": {
            "agent_name": "test-agent",
            "custom_tools": ["mock_tool1", "mock_tool2"],
            "base_url": "https://test.com",
            "api_key": "test-key"
        },
        "expected": {
            "agent_name": "test-agent",
            "custom_tools": ["mock_tool1", "mock_tool2"]
        }
    },
    {
        "id": "with_model_context",
        "params": {
            "agent_name": "test-agent",
            "model_context": "mock_context",
            "base_url": "https://test.com",
            "api_key": "test-key"
        },
        "expected": {
            "agent_name": "test-agent",
            "model_context": "mock_context"
        }
    }
]

# Test data for API URL construction
API_URL_DATA = [
    {
        "id": "basic_url",
        "agent_name": "my-agent",
        "base_url": "https://api.test.com",
        "expected_url": "https://api.test.com/v1/agents/my-agent"
    },
    {
        "id": "url_with_trailing_slash",
        "agent_name": "my-agent",
        "base_url": "https://api.test.com/",
        "expected_url": "https://api.test.com/v1/agents/my-agent"
    },
    {
        "id": "complex_agent_name",
        "agent_name": "my-complex-agent-name",
        "base_url": "https://api.test.com",
        "expected_url": "https://api.test.com/v1/agents/my-complex-agent-name"
    }
]

# Test data for sync functionality
SYNC_TEST_DATA = [
    {
        "id": "sync_disabled",
        "config": MINIMAL_CONFIG,
        "time_elapsed": 100,
        "should_sync": False
    },
    {
        "id": "sync_enabled_interval_not_passed",
        "config": CONFIG_WITH_SYNC,
        "time_elapsed": 5,
        "should_sync": False
    },
    {
        "id": "sync_enabled_interval_passed",
        "config": CONFIG_WITH_SYNC,
        "time_elapsed": 15,
        "should_sync": True
    }
]




"""Test data for autogen sessions function parametrized tests."""

from autogen_core import FunctionCall
from autogen_core.models import (
    UserMessage,
    SystemMessage,
    AssistantMessage,
    FunctionExecutionResult,
    FunctionExecutionResultMessage,
)

# Sample message for testing (used in test_sessions.py)
SAMPLE_USER_MESSAGE = UserMessage(content="Hi my name is Ashok", source="user")

# Test data for _autogen_to_flotorch_event method
AUTOGEN_TO_FLOTORCH_TEST_DATA = [
    (
        "user_message_string_content",
        UserMessage(content="Hi my name is Ashok", source="user"),
        ("user", {"message": "Hi my name is Ashok"})
    ),
    (
        "system_message_string_content",
        SystemMessage(content="You are a helpful assistant", source="system"),
        ("system", {"message": "You are a helpful assistant"})
    ),
    (
        "assistant_message_string_content",
        AssistantMessage(content="Hello, Ashok! How can I assist you today?",
                         source="FlotorchAssistant"),
        ("assistant", {"message": "Hello, Ashok! How can I assist you today?"})
    ),
    (
        "assistant_message_with_tool_calls",
        AssistantMessage(
            content=[FunctionCall(
                id="call_123",
                name="calculate",
                arguments='{"expression": "21+213"}'
            )],
            source="FlotorchAssistant"
        ),
        ("assistant", {
            "tool_calls": [{
                "id": "call_123",
                "function": {
                    "name": "calculate",
                    "arguments": '{"expression": "21+213"}'
                }
            }]
        })
    ),
    (
        "function_execution_result_message",
        FunctionExecutionResultMessage(
            content=[FunctionExecutionResult(
                call_id="call_123",
                name="calculate",
                content="234",
                is_error=False
            )],
            source="tool"
        ),
        ("tool", {
            "tool_outputs": [{
                "tool_call_id": "call_123",
                "name": "calculate",
                "output": "234",
                "is_error": False
            }]
        })
    ),
]

# Test data for _flotorch_to_autogen_message method
FLOTORCH_TO_AUTOGEN_TEST_DATA = [
    (
        "user_event_with_message",
        {
            "author": "user",
            "content": {"message": "Hi my name is Ashok"},
            "timestamp": 1640995200000
        },
        UserMessage(content="Hi my name is Ashok", source="user")
    ),
    (
        "system_event_with_message",
        {
            "author": "system",
            "content": {"message": "You are a helpful assistant"},
            "timestamp": 1640995200000
        },
        SystemMessage(content="You are a helpful assistant", source="system")
    ),
    (
        "assistant_event_with_message",
        {
            "author": "assistant",
            "content": {"message": "Hello, Ashok! How can I assist you today?"},
            "timestamp": 1640995200000
        },
        AssistantMessage(content="Hello, Ashok! How can I assist you today?",
                         source="assistant")
    ),
    (
        "assistant_event_with_tool_calls",
        {
            "author": "assistant",
            "content": {
                "tool_calls": [{
                    "id": "call_123",
                    "function": {
                        "name": "calculate",
                        "arguments": '{"expression": "21+213"}'
                    }
                }]
            },
            "timestamp": 1640995200000
        },
        AssistantMessage(
            content=[FunctionCall(id="call_123", name="calculate",
                                  arguments='{"expression": "21+213"}')],
            source="assistant"
        )
    ),
    (
        "tool_event_with_outputs",
        {
            "author": "tool",
            "content": {
                "tool_outputs": [{
                    "tool_call_id": "call_123",
                    "name": "calculate",
                    "output": "234",
                    "is_error": False
                }]
            },
            "timestamp": 1640995200000
        },
        FunctionExecutionResultMessage(
            content=[FunctionExecutionResult(
                call_id="call_123",
                name="calculate",
                content="234",
                is_error=False
            )],
            source="tool"
        )
    ),
    (
        "event_without_author",
        {
            "content": {"message": "No author"},
            "timestamp": 1640995200000
        },
        None
    ),
    (
        "event_with_empty_content",
        {
            "author": "user",
            "content": {},
            "timestamp": 1640995200000
        },
        UserMessage(content="", source="user")
    ),
    (
        "unknown_author_event",
        {
            "author": "unknown",
            "content": {"message": "Unknown author"},
            "timestamp": 1640995200000
        },
        None
    ),
]

# Test data for add_message method
ADD_MESSAGE_TEST_DATA = [
    (
        "valid_user_message",
        UserMessage(content="Hi my name is Ashok", source="user"),
        False,
        None
    ),
    (
        "valid_assistant_message",
        AssistantMessage(content="Hello, Ashok! How can I assist you today?",
                         source="FlotorchAssistant"),
        False,
        None
    ),
    (
        "message_with_tool_calls",
        AssistantMessage(
            content=[FunctionCall(
                id="call_123",
                name="calculate",
                arguments='{"expression": "21+213"}',
            )],
            source="FlotorchAssistant"
        ),
        False,
        None
    ),
    (
        "message_without_author_conversion",
        UserMessage(content="valid message", source="user"),
        False,
        None
    ),
    (
        "api_failure_scenario",
        UserMessage(content="test message", source="user"),
        True,
        Exception
    ),
]

# Test data for get_messages method
GET_MESSAGES_TEST_DATA = [
    (
        "empty_cache",
        [],
        50,
        0
    ),
    (
        "cache_with_few_messages",
        [
            UserMessage(content="Message 1", source="user"),
            AssistantMessage(content="Response 1", source="assistant")
        ],
        50,
        2
    ),
    (
        "cache_exceeding_limit",
        [
            UserMessage(content=f"Message {i}", source="user")
            for i in range(60)
        ],
        50,
        50
    ),
    (
        "cache_exactly_at_limit",
        [
            UserMessage(content=f"Message {i}", source="user")
            for i in range(50)
        ],
        50,
        50
    ),
]

# Test data for initialization scenarios
INITIALIZATION_TEST_DATA = [
    (
        "new_session_creation",
        None,
        False,
        None
    ),
    (
        "existing_session_loading",
        "existing-session-123",
        False,
        None
    ),
    (
        "session_creation_failure",
        None,
        True,
        RuntimeError
    ),
    (
        "session_loading_failure",
        "invalid-session-123",
        True,
        RuntimeError
    ),
]

# Test data for state management
STATE_MANAGEMENT_TEST_DATA = [
    (
        "save_state_with_uid",
        {"uid": "test-session-123", "base_url": "https://test.com"},
        {
            "uid": "test-session-123",
            "base_url": "https://test.com",
            "provider": "flotorch.autogen.sessions.FlotorchAutogenSession"}
    ),
    (
        "save_state_without_uid",
        {"uid": None, "base_url": "https://test.com"},
        {
            "uid": None,
            "base_url": "https://test.com", 
            "provider": "flotorch.autogen.sessions.FlotorchAutogenSession"}
    ),
    (
        "load_state_valid",
        {"uid": "test-session-123", "base_url": "https://test.com"},
        False,
        None
    ),
    (
        "load_state_invalid",
        {"invalid": "data"},
        True,
        ValueError
    ),
]



"""Tests for FlotorchStore class."""

import pytest
import json
from unittest.mock import patch
from datetime import datetime
from langgraph.store.base import GetOp, PutOp, SearchOp

from flotorch.langgraph.tests.test_data.memory_test_data import (
    GET_TEST_DATA,
    PUT_TEST_DATA,
    SEARCH_TEST_DATA,
    UUID_VALIDATION_TEST_DATA,
    DATA_CONVERSION_TEST_DATA
)


class TestStoreInit:
    """Test initialization of FlotorchStore."""

    @patch('flotorch.langgraph.memory.FlotorchMemory')
    @patch('flotorch.langgraph.memory.FlotorchAsyncMemory')
    def test_init_with_all_parameters(self, mock_async_memory, mock_memory):
        """Test initialization with all custom parameters."""
        from flotorch.langgraph.memory import FlotorchStore
        instance = FlotorchStore(
            api_key="custom-key",
            base_url="https://custom.flotorch.com",
            provider_name="custom_provider",
            userId="user_123",
            agentId="agent_456",
            appId="app_789",
            sessionId="session_abc"
        )
        assert (instance.api_key == "custom-key" and
                instance.base_url == "https://custom.flotorch.com" and
                instance.provider_name == "custom_provider" and
                instance.userId == "user_123" and
                instance.agentId == "agent_456" and
                instance.appId == "app_789" and
                instance.sessionId == "session_abc")
        mock_memory.assert_called_once_with(
            "custom-key", "https://custom.flotorch.com", "custom_provider")
        mock_async_memory.assert_called_once_with(
            "custom-key", "https://custom.flotorch.com", "custom_provider")

    @patch('flotorch.langgraph.memory.FlotorchMemory')
    @patch('flotorch.langgraph.memory.FlotorchAsyncMemory')
    def test_init_with_minimal_parameters(
        self, mock_async_memory, mock_memory
    ):
        """Test initialization with only required parameters."""
        from flotorch.langgraph.memory import FlotorchStore
        instance = FlotorchStore(
            api_key="test-key", base_url="https://test.flotorch.com",
            provider_name="test_provider"
        )
        assert (instance.api_key == "test-key" and
                instance.base_url == "https://test.flotorch.com" and
                instance.provider_name == "test_provider" and
                instance.userId is None and instance.agentId is None and
                instance.appId is None and instance.sessionId is None and
                instance.supports_ttl is False)


class TestNamespaceMapping:
    """Test namespace to gateway IDs mapping functionality."""

    def test_namespace_single_and_two_levels(
        self, store_instance_no_defaults
    ):
        """Test single and two level namespace mapping."""
        result = store_instance_no_defaults._namespace_to_gateway_ids(
            ("users",))
        assert result.get("appId") == "users" and "userId" not in result
        result = store_instance_no_defaults._namespace_to_gateway_ids(
            ("users", "alice"))
        assert (result.get("appId") == "users" and
                result.get("userId") == "alice")

    def test_namespace_with_class_defaults(self, store_instance):
        """Test empty namespace uses class defaults and overrides."""
        result = store_instance._namespace_to_gateway_ids(())
        assert (result["appId"] == "test_app" and
                result["userId"] == "test_user" and
                result["agentId"] == "test_agent" and
                result["sessionId"] == "test_session")
        result = store_instance._namespace_to_gateway_ids(
            ("custom_app", "custom_user"))
        assert (result["appId"] == "custom_app" and
                result["userId"] == "custom_user" and
                result["agentId"] == "test_agent" and
                result["sessionId"] == "test_session")


class TestUUIDValidation:
    """Test UUID validation functionality."""

    @pytest.mark.parametrize(
        "test_name,key,expected_result", UUID_VALIDATION_TEST_DATA
    )
    def test_is_memory_id(
        self, store_instance, test_name, key, expected_result
    ):
        """Test UUID validation with various key formats."""
        assert store_instance._is_memory_id(key) == expected_result


class TestDataConversion:
    """Test data conversion between memory and LangGraph formats."""

    @pytest.mark.parametrize(
        "test_name,memory_data,expected_value", DATA_CONVERSION_TEST_DATA
    )
    def test_memory_data_to_item(
        self,
        store_instance,
        test_name,
        memory_data,
        expected_value
    ):
        """Test conversion from memory data to LangGraph Item."""
        item = store_instance._memory_data_to_item(
            memory_data, ("users", "alice"), "test_key")
        assert (item.value == expected_value and item.key == "test_key" and
                item.namespace == ("users", "alice") and
                isinstance(item.created_at, datetime) and
                isinstance(item.updated_at, datetime))

    def test_memory_data_to_search_item(self, store_instance):
        """Test conversion from memory data to LangGraph SearchItem."""
        memory_data = {
            "id": "mem-123", "content": '{"test": "data"}',
            "createdAt": "2024-01-01T00:00:00",
            "updatedAt": "2024-01-01T00:00:00", "importance": 0.8,
            "metadata": {
                "langgraph_namespace": ["users", "alice"],
                "langgraph_key": "test_key"}
        }
        search_item = store_instance._memory_data_to_search_item(memory_data)
        assert (search_item.namespace == ("users", "alice") and
                search_item.key == "test_key" and
                search_item.value == {"test": "data"} and
                search_item.score == 0.8)
        memory_data["metadata"] = {}
        op = SearchOp(namespace_prefix=("default_ns",), query="test")
        search_item = store_instance._memory_data_to_search_item(
            memory_data, op)
        assert search_item.namespace == ("default_ns",)

    def test_response_to_item_conversion(self, store_instance):
        """Test conversion from API response to Item."""
        response = {
            "data": {
                "id": "mem-123",
                "content": '{"key": "value"}',
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00"
            }
        }
        item = store_instance._response_to_item(
            response, ("users",), "test_key")
        assert (item is not None and item.value == {"key": "value"} and
                item.key == "test_key")
        assert store_instance._response_to_item(
            {}, ("users",), "test_key") is None


class TestGetOperations:
    """Test GET operation functionality."""

    @pytest.mark.parametrize(
        "test_name,key,namespace,api_response,should_succeed,"
        "should_fail,expected_exception", GET_TEST_DATA
    )
    def test_handle_get_with_various_scenarios(
        self,
        store_instance,
        mock_memory_client,
        test_name,
        key,
        namespace,
        api_response,
        should_succeed,
        should_fail,
        expected_exception
    ):
        """Test GET operation with parametrized scenarios."""
        if should_fail:
            mock_memory_client.get.side_effect = Exception("API Error")
            mock_memory_client.search.side_effect = Exception("API Error")
        else:
            if store_instance._is_memory_id(key):
                mock_memory_client.get.return_value = api_response
            else:
                mock_memory_client.search.return_value = api_response
        op = GetOp(namespace=namespace, key=key)
        result = store_instance._handle_get(op)
        if should_succeed:
            assert (result is not None and result.key == key and
                    result.namespace == namespace)
        else:
            assert result is None

    def test_get_by_uuid_vs_key(self, store_instance, mock_memory_client):
        """Test UUID keys use get() and non-UUID keys use search()."""
        uuid_key = "123e4567-e89b-12d3-a456-426614174000"
        mock_memory_client.get.return_value = {
            "data": {
                "id": uuid_key,
                "content": '{"test": "data"}',
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00"
            }
        }
        op = GetOp(namespace=("users",), key=uuid_key)
        result = store_instance._handle_get(op)
        mock_memory_client.get.assert_called_once_with(uuid_key)
        assert result is not None
        mock_memory_client.reset_mock()
        mock_memory_client.search.return_value = {
            "data": [{
                "id": "mem-123", "content": "Test content",
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00",
                "metadata": {"tags": ["test_key"], "langgraph_key": "test_key"}
            }]
        }
        op = GetOp(namespace=("users",), key="test_key")
        result = store_instance._handle_get(op)
        mock_memory_client.search.assert_called_once()
        call_kwargs = mock_memory_client.search.call_args.kwargs
        assert (call_kwargs["appId"] == "users" and
                call_kwargs["limit"] == 1 and result is not None)

    def test_get_aggregates_multiple_results(
        self, store_instance, mock_memory_client
    ):
        """Test get aggregates content from multiple results."""
        mock_memory_client.search.return_value = {
            "data": [
                {"id": "mem-1", "content": "First part",
                 "createdAt": "2024-01-01T00:00:00",
                 "updatedAt": "2024-01-01T00:00:00",
                 "metadata": {"tags": ["test_key"]}},
                {"id": "mem-2", "content": "Second part",
                 "createdAt": "2024-01-01T00:00:00",
                 "updatedAt": "2024-01-01T00:00:00",
                 "metadata": {"tags": ["test_key"]}}]
        }
        op = GetOp(namespace=("users",), key="test_key")
        result = store_instance._handle_get(op)
        assert (result is not None and
                "First part" in result.value.get("text", "") and
                "Second part" in result.value.get("text", ""))

    @pytest.mark.asyncio
    async def test_handle_aget_async(
        self, store_instance, mock_async_memory_client
    ):
        """Test asynchronous GET operation."""
        uuid_key = "123e4567-e89b-12d3-a456-426614174000"
        mock_async_memory_client.get.return_value = {
            "data": {
                "id": uuid_key,
                "content": '{"test": "async data"}',
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00"
            }
        }
        op = GetOp(namespace=("users",), key=uuid_key)
        result = await store_instance._handle_aget(op)
        mock_async_memory_client.get.assert_called_once_with(uuid_key)
        assert (result is not None and
                result.value == {"test": "async data"})


class TestPutOperations:
    """Test PUT operation functionality."""

    @pytest.mark.parametrize(
        "test_name,key,namespace,value,delete_response,is_delete,"
        "should_fail,expected_exception", PUT_TEST_DATA
    )
    def test_handle_put_with_various_scenarios(
        self,
        store_instance,
        mock_memory_client,
        test_name,
        key,
        namespace,
        value,
        delete_response,
        is_delete,
        should_fail,
        expected_exception
    ):
        """Test PUT operation with parametrized scenarios."""
        if should_fail:
            mock_memory_client.add.side_effect = Exception("API Error")
            mock_memory_client.delete.side_effect = Exception("API Error")
        else:
            if is_delete and not store_instance._is_memory_id(key):
                mock_memory_client.search.return_value = delete_response
        op = PutOp(namespace=namespace, key=key, value=value)
        store_instance._handle_put(op)
        if not should_fail:
            if is_delete:
                if store_instance._is_memory_id(key):
                    mock_memory_client.delete.assert_called_once_with(key)
            else:
                mock_memory_client.add.assert_called_once()

    def test_put_stores_with_correct_metadata(
        self, store_instance, mock_memory_client
    ):
        """Test put stores data with correct metadata."""
        value = {"preference": "dark_mode", "language": "en"}
        op = PutOp(
            namespace=("users", "alice"), key="settings", value=value)
        store_instance._handle_put(op)
        call_kwargs = mock_memory_client.add.call_args.kwargs
        assert (call_kwargs["appId"] == "users" and
                call_kwargs["userId"] == "alice" and
                "messages" in call_kwargs and
                call_kwargs["metadata"]["source"] == "langgraph_store" and
                call_kwargs["metadata"]["tags"] == ["settings"])
        message_content = call_kwargs["messages"][0]["content"]
        assert json.loads(message_content) == value

    def test_delete_operations(self, store_instance, mock_memory_client):
        """Test delete by UUID and by key."""
        uuid_key = "123e4567-e89b-12d3-a456-426614174000"
        op = PutOp(namespace=("users",), key=uuid_key, value=None)
        store_instance._handle_put(op)
        mock_memory_client.delete.assert_called_once_with(uuid_key)
        mock_memory_client.reset_mock()
        mock_memory_client.search.return_value = {
            "data": [{"id": "mem-to-delete"}]}
        op = PutOp(namespace=("users",), key="old_setting", value=None)
        store_instance._handle_put(op)
        mock_memory_client.search.assert_called_once()
        mock_memory_client.delete.assert_called_once_with("mem-to-delete")

    @pytest.mark.asyncio
    async def test_handle_aput_operations(
        self, store_instance, mock_async_memory_client
    ):
        """Test asynchronous PUT and DELETE operations."""
        value = {"test": "async value"}
        op = PutOp(namespace=("users",), key="test_key", value=value)
        await store_instance._handle_aput(op)
        mock_async_memory_client.add.assert_called_once()
        mock_async_memory_client.reset_mock()
        uuid_key = "123e4567-e89b-12d3-a456-426614174000"
        op = PutOp(namespace=("users",), key=uuid_key, value=None)
        await store_instance._handle_aput(op)
        mock_async_memory_client.delete.assert_called_once_with(uuid_key)


class TestSearchOperations:
    """Test SEARCH operation functionality."""

    @pytest.mark.parametrize(
        "test_name,query,namespace,limit,api_response,expected_count,"
        "should_fail,expected_exception", SEARCH_TEST_DATA
    )
    def test_handle_search_with_various_scenarios(
        self,
        store_instance,
        mock_memory_client,
        test_name,
        query,
        namespace,
        limit,
        api_response,
        expected_count,
        should_fail,
        expected_exception
    ):
        """Test SEARCH operation with parametrized scenarios."""
        if should_fail:
            mock_memory_client.search.side_effect = Exception("API Error")
        else:
            mock_memory_client.search.return_value = api_response
        op = SearchOp(namespace_prefix=namespace, query=query, limit=limit)
        results = store_instance._handle_search(op)
        assert len(results) == expected_count
        if expected_count > 0:
            for item in results:
                assert (hasattr(item, "namespace") and
                        hasattr(item, "key") and hasattr(item, "value") and
                        hasattr(item, "score"))

    def test_search_query_and_limits(
        self, store_instance, mock_memory_client
    ):
        """Test search uses semantic query and respects limits."""
        mock_memory_client.search.return_value = {"data": []}
        op = SearchOp(
            namespace_prefix=("users",), query="user preferences", limit=10)
        store_instance._handle_search(op)
        call_kwargs = mock_memory_client.search.call_args.kwargs
        assert (call_kwargs["query"] == "user preferences" and
                call_kwargs["limit"] == 10 and
                call_kwargs["appId"] == "users")
        mock_memory_client.reset_mock()
        mock_memory_client.search.return_value = {
            "data": [
                {"id": f"mem-{i}", "content": f"Content {i}",
                 "createdAt": "2024-01-01T00:00:00",
                 "updatedAt": "2024-01-01T00:00:00",
                 "importance": 0.5, "metadata": {}} for i in range(10)]
        }
        op = SearchOp(namespace_prefix=("users",), query="test", limit=3)
        results = store_instance._handle_search(op)
        assert len(results) == 3
        mock_memory_client.reset_mock()
        mock_memory_client.search.return_value = {"data": []}
        op = SearchOp(namespace_prefix=("users",), query="test", limit=None)
        store_instance._handle_search(op)
        call_kwargs = mock_memory_client.search.call_args.kwargs
        assert call_kwargs["limit"] == 20

    @pytest.mark.asyncio
    async def test_handle_asearch_async(
        self, store_instance, mock_async_memory_client
    ):
        """Test asynchronous SEARCH operation."""
        mock_async_memory_client.search.return_value = {
            "data": [{
                "id": "mem-1",
                "content": '{"test": "data"}',
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00",
                "importance": 0.7,
                "metadata": {}
            }]
        }
        op = SearchOp(
            namespace_prefix=("users",), query="test query", limit=5)
        results = await store_instance._handle_asearch(op)
        mock_async_memory_client.search.assert_called_once()
        assert len(results) == 1


class TestBatchOperations:
    """Test BATCH operation functionality."""

    def test_batch_multiple_operations(
        self, store_instance, mock_memory_client
    ):
        """Test batch processing of multiple operations."""
        mock_memory_client.get.return_value = {
            "data": {
                "id": "mem-123", "content": '{"test": "data"}',
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00"
            }
        }
        ops = [
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174000"),
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174001"),
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174002"),
        ]
        results = store_instance.batch(ops)
        assert (len(results) == 3 and
                mock_memory_client.get.call_count == 3)

    def test_batch_mixed_operation_types(
        self, store_instance, mock_memory_client
    ):
        """Test batch processing with mixed operation types."""
        mock_memory_client.get.return_value = {
            "data": {
                "id": "mem-123", "content": '{"test": "data"}',
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00"
            }
        }
        mock_memory_client.search.return_value = {"data": []}
        ops = [
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174000"),
            PutOp(namespace=("users",), key="key1",
                  value={"data": "test"}),
            SearchOp(namespace_prefix=("users",), query="test", limit=5),
        ]
        results = store_instance.batch(ops)
        assert (len(results) == 3 and
                mock_memory_client.get.call_count == 1 and
                mock_memory_client.add.call_count == 1 and
                mock_memory_client.search.call_count == 1)

    def test_batch_handles_individual_failures(
        self, store_instance, mock_memory_client
    ):
        """Test batch continues processing when operations fail."""
        mock_memory_client.get.side_effect = [
            {"data": {"id": "mem-1", "content": '{"test": "data"}',
                     "createdAt": "2024-01-01T00:00:00",
                     "updatedAt": "2024-01-01T00:00:00"}},
            Exception("API Error"),
            {"data": {"id": "mem-3", "content": '{"test": "data"}',
                     "createdAt": "2024-01-01T00:00:00",
                     "updatedAt": "2024-01-01T00:00:00"}},
        ]
        ops = [
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174000"),
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174001"),
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174002"),
        ]
        results = store_instance.batch(ops)
        assert (len(results) == 3 and results[0] is not None and
                results[1] is None and results[2] is not None)

    @pytest.mark.asyncio
    async def test_abatch_async_operations(
        self, store_instance, mock_async_memory_client
    ):
        """Test asynchronous batch processing."""
        mock_async_memory_client.get.return_value = {
            "data": {
                "id": "mem-123", "content": '{"test": "async data"}',
                "createdAt": "2024-01-01T00:00:00",
                "updatedAt": "2024-01-01T00:00:00"
            }
        }
        mock_async_memory_client.search.return_value = {"data": []}
        ops = [
            GetOp(namespace=("users",),
                  key="123e4567-e89b-12d3-a456-426614174000"),
            PutOp(namespace=("users",), key="key1",
                  value={"data": "test"}),
            SearchOp(namespace_prefix=("users",), query="test", limit=5),
        ]
        results = await store_instance.abatch(ops)
        assert len(results) == 3


class TestEdgeCases:
    """Test edge cases and error handling."""

    def test_malformed_json_content_handling(self, store_instance):
        """Test that malformed JSON is handled gracefully."""
        memory_data = {
            "id": "mem-bad", "content": "{invalid json}",
            "createdAt": "2024-01-01T00:00:00",
            "updatedAt": "2024-01-01T00:00:00"
        }
        item = store_instance._memory_data_to_item(
            memory_data, ("users",), "test_key")
        assert isinstance(item.value, dict) and "text" in item.value

    def test_unicode_content_handling(
        self, store_instance, mock_memory_client
    ):
        """Test proper handling of Unicode content."""
        unicode_value = {"message": "Hello World", "language": "English"}
        op = PutOp(
            namespace=("users",), key="unicode_key", value=unicode_value)
        store_instance._handle_put(op)
        call_kwargs = mock_memory_client.add.call_args.kwargs
        content = call_kwargs["messages"][0]["content"]
        parsed = json.loads(content)
        assert (parsed["message"] == "Hello World" and
                parsed["language"] == "English")

    def test_empty_namespace_handling(self, store_instance_no_defaults):
        """Test operations with empty namespace."""
        result = store_instance_no_defaults._namespace_to_gateway_ids(())
        assert isinstance(result, dict)

    def test_batch_with_list_namespaces_operation(self, store_instance):
        """Test batch handles ListNamespacesOp."""
        from langgraph.store.base import ListNamespacesOp
        ops = [ListNamespacesOp(match_conditions=[], max_depth=None)]
        results = store_instance.batch(ops)
        assert len(results) == 1 and results[0] is None


class TestConvenienceFunction:
    """Test the convenience function for creating FlotorchStore."""

    @patch('flotorch.langgraph.memory.FlotorchMemory')
    @patch('flotorch.langgraph.memory.FlotorchAsyncMemory')
    def test_create_flotorch_store(self, mock_async_memory, mock_memory):
        """Test the create_flotorch_store convenience function."""
        from flotorch.langgraph.memory import create_flotorch_store
        store = create_flotorch_store(
            api_key="test-key", base_url="https://test.com",
            provider_name="test_provider", userId="user_123",
            agentId="agent_456", appId="app_789", sessionId="session_abc"
        )
        assert (store.api_key == "test-key" and
                store.base_url == "https://test.com" and
                store.provider_name == "test_provider" and
                store.userId == "user_123" and
                store.agentId == "agent_456" and
                store.appId == "app_789" and
                store.sessionId == "session_abc")



"""Test data for Strands Memory tests based on actual runtime execution."""

# Memory ADD operation test data
MEMORY_ADD_TEST_DATA = [
    (
        "add_simple_text",
        {"action": "add", "content": "User's name is Madhu."},
        {
            "object": "agent.memory.list",
            "data": [
                {
                    "id": "2582aa4a-d728-4cb3-b32c-a688c336c24e",
                    "content": "User Name is Madhu"
                }
            ]
        },
        "Memory stored successfully. ID: unknown",
        False,  # should_fail
        None    # expected_exception
    ),
    (
        "add_complex_info",
        {"action": "add", "content": "User scored 1000 in a test in 2010"},
        {
            "object": "agent.memory.list",
            "data": [
                {
                    "id": "81aedc81-40b9-4435-a7fd-e343b8c046cc",
                    "content": "User Scored 1000 in a test in 2010"
                }
            ]
        },
        "Memory stored successfully. ID: unknown",
        False,
        None
    ),
    (
        "add_missing_content",
        {"action": "add"},
        None,
        "Error: content is required for add action",
        True,  # should_fail
        ValueError
    ),
]

# Memory SEARCH operation test data
MEMORY_SEARCH_TEST_DATA = [
    (
        "search_user_name",
        {"action": "search", "query": "User's name"},
        {
            "object": "agent.memory.list",
            "data": [
                {
                    "id": "2582aa4a-d728-4cb3-b32c-a688c336c24e",
                    "content": "User Name is Madhu",
                    "userId": "user-id-15",
                    "appId": "app-id-15",
                    "metadata": {"tags": [], "source": "strands", "importance": 0.5},
                    "createdAt": "2025-09-24T05:35:12.846Z",
                    "updatedAt": "2025-10-13T08:00:06.005Z",
                    "score": 0.8169173466955095
                },
                {
                    "id": "81aedc81-40b9-4435-a7fd-e343b8c046cc",
                    "content": "User Scored 1000 in a test in 2010",
                    "userId": "user-id-15",
                    "appId": "app-id-15",
                    "metadata": {"tags": [], "source": "strands", "importance": 0.5},
                    "createdAt": "2025-09-30T11:30:55.032Z",
                    "updatedAt": "2025-09-30T11:30:55.092Z",
                    "score": 0.4680134290720004
                }
            ]
        },
        "Found 2 relevant memories:\n1. User Name is Madhu\n2. User Scored 1000 in a test in 2010\n",
        False,
        None
    ),
    (
        "search_no_results",
        {"action": "search", "query": "nonexistent information"},
        {
            "object": "agent.memory.list",
            "data": []
        },
        "No relevant memories found.",
        False,
        None
    ),
    (
        "search_missing_query",
        {"action": "search"},
        None,
        "Error: query is required for search action",
        True,
        ValueError
    ),
]

# Memory LIST operation test data
MEMORY_LIST_TEST_DATA = [
    (
        "list_all_memories",
        {"action": "list"},
        {
            "object": "agent.memory.list",
            "data": [
                {
                    "id": "2582aa4a-d728-4cb3-b32c-a688c336c24e",
                    "content": "User Name is Madhu",
                    "userId": "user-id-15",
                    "appId": "app-id-15",
                    "metadata": {"tags": [], "source": "strands", "importance": 0.5},
                    "createdAt": "2025-09-24T05:35:12.846Z",
                    "updatedAt": "2025-10-13T08:00:06.005Z"
                },
                {
                    "id": "81aedc81-40b9-4435-a7fd-e343b8c046cc",
                    "content": "User Scored 1000 in a test in 2010",
                    "userId": "user-id-15",
                    "appId": "app-id-15",
                    "metadata": {"tags": [], "source": "strands", "importance": 0.5},
                    "createdAt": "2025-09-30T11:30:55.032Z",
                    "updatedAt": "2025-09-30T11:30:55.092Z"
                }
            ]
        },
        "Total memories: 2\n1. [2582aa4a-d728-4cb3-b32c-a688c336c24e] User Name is Madhu...\n2. [81aedc81-40b9-4435-a7fd-e343b8c046cc] User Scored 1000 in a test in 2010...\n",
        False,
        None
    ),
    (
        "list_empty_memories",
        {"action": "list"},
        {
            "object": "agent.memory.list",
            "data": []
        },
        "No memories found.",
        False,
        None
    ),
]

# Invalid action test data
INVALID_ACTION_TEST_DATA = [
    (
        "invalid_action",
        {"action": "delete"},
        None,
        "Error: Invalid action: delete",
        True,
        ValueError
    ),
    (
        "missing_action",
        {},
        None,
        "Error: action parameter is required",
        True,
        ValueError
    ),
]

# Tool Result format test data
TOOL_RESULT_FORMAT_TEST_DATA = [
    (
        "success_result_format",
        {
            "toolUseId": "call_test_123",
            "status": "success",
            "content": [{"text": "Memory stored successfully. ID: unknown"}]
        },
        "success",
        "Memory stored successfully. ID: unknown"
    ),
    (
        "error_result_format",
        {
            "toolUseId": "call_test_456",
            "status": "error",
            "content": [{"text": "Error: content is required for add action"}]
        },
        "error",
        "Error: content is required for add action"
    ),
]




"""
Flotorch Strands integration module.

This module provides integration between Flotorch SDK and Strands Agents framework,
allowing you to use Flotorch's LLM gateway with Strands agents.
"""

from .llm import FlotorchStrandsModel

__all__ = ["FlotorchStrandsModel"]

